#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
import gensim
import numpy as np
import pandas as pd
from attr import dataclass

from nltk.tokenize import word_tokenize
from logai.algorithms.algo_interfaces import VectorizationAlgo
from logai.config_interfaces import Config
from logai.algorithms.factory import factory


@dataclass
class Word2VecParams(Config):
    """
    Configuration of Word2Vec vectorization parameters. For more details on the parameters see 
    https://radimrehurek.com/gensim/models/word2vec.html.

    :param max_token_len: The maximum length of tokens.
    :param min_count: Ignores all words with total frequency lower than this.
    :param vector_size: Dimensionality of the feature vectors.
    :param window: The maximum distance between the current and predicted word within a sentence.
    """

    max_token_len: int = 100
    min_count: int = 1
    vector_size: int = 3
    window: int = 3


@factory.register("vectorization", "word2vec", Word2VecParams)
class Word2Vec(VectorizationAlgo):
    """
    Word2Vec algorithm for converting raw log data into word2vec vectors. This is a wrapper class for the Word2Vec
    model from gensim library https://radimrehurek.com/gensim/models/word2vec.html

    :param max_token_len: The max token length to vectorize, longer sentences will be chopped.
    """

    def __init__(self, params: Word2VecParams):
        self.params = params
        self.model = None

    def fit(self, loglines: pd.Series):
        """
        Fits a Word2Vec model.

        :param loglines: Parsed loglines.
        """
        max_token_len = self.params.max_token_len

        doc = []
        for sentence in loglines:
            token_list = sentence.split(" ")[:max_token_len]
            for tk in token_list:
                if tk != "*":
                    doc.append(word_tokenize(tk.lower()))

        # Use Word2Vec for vectorization
        self.model = gensim.models.Word2Vec(
            doc,
            min_count=self.params.min_count,
            vector_size=self.params.vector_size,
            window=self.params.window,
        )

    def transform(self, loglines: pd.Series) -> pd.Series:
        """
        Transforms input loglines to log vectors.

        :param loglines: The input loglines.
        :return: The transformed log vectors.
        """
        log_vectors = []
        max_len = 0
        for ll in loglines:
            token_list = ll.split(" ")

            log_vector = []

            token_list = token_list[: self.params.max_token_len]

            max_len = max(max_len, len(token_list))
            for tk in token_list:
                if tk == "*":
                    continue
                log_vector.append(self.model.wv[word_tokenize(tk.lower())][0])
            log_vectors.append(np.array(log_vector).flatten())
        log_vector_series = pd.Series(log_vectors, index=loglines.index)
        return log_vector_series

    def summary(self):
        """
        Generates model summary.
        """
        return self.model.summary()



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
from .preprocessor import Preprocessor, PreprocessorConfig
import pandas as pd
import re
from logai.dataloader.data_model import LogRecordObject
from logai.utils import constants


class OpenSetPreprocessor(Preprocessor):
    """Preprocessor class for Open log datasets.
        
    :param config: A config object specifying parameters of log preprocessing for open log datasets.
    """
    def __init__(self, config: PreprocessorConfig):
        super().__init__(config)
        self.config = config

    def _get_ids(self, logrecord: LogRecordObject) -> pd.Series:
        return None

    def _get_labels(self, logrecord: LogRecordObject) -> pd.Series:
        return None

    def _format_ids(self, data_id: pd.Series):
        if type(list(data_id)[0]) == str:
            id_to_serial_id_map = {k: i for i, k in enumerate(set(data_id))}
            data_id = data_id.apply(lambda x: id_to_serial_id_map[x])
        return data_id

    def clean_log(self, logrecord: LogRecordObject) -> LogRecordObject:
        """Preprocessing cleaning of logrecord object creating from open log datasets.
        
        :param logrecord: A log record object containing the raw log data from open datasets.
        :return: The cleaned logrecord object.
        """
        preprocessed_loglines, custom_patterns = super().clean_log(
            logrecord.body[constants.LOGLINE_NAME]
        )
        for pattern in self.config.custom_replace_list:
            pattern = pattern[1]
            regex = r"((" + pattern + ")[ /=]*)+"
            value = pattern
            preprocessed_loglines = preprocessed_loglines.replace(
                to_replace=regex, value=value, regex=True
            )
        preprocessed_loglines = preprocessed_loglines.apply(
            lambda x: re.sub(" +", " ", x.replace("*", ""))
        )

        logrecord.body = pd.DataFrame({constants.LOGLINE_NAME: preprocessed_loglines})
        logrecord.body = logrecord.body.join(pd.DataFrame(custom_patterns))
        preprocessed_ids = self._get_ids(logrecord)
        preprocessed_ids = self._format_ids(preprocessed_ids)

        logrecord.span_id = pd.DataFrame({constants.SPAN_ID: preprocessed_ids})
        logrecord.labels = pd.DataFrame({constants.LABELS: self._get_labels(logrecord)})
        return logrecord



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
from torch import nn
from logai.algorithms.nn_model.forecast_nn.base_nn import (
    ForecastBasedNN,
    ForecastBasedNNParams,
)
from attr import dataclass
from logai.algorithms.vectorization_algo.forecast_nn import ForecastNNVectorizedDataset


@dataclass
class TransformerParams(ForecastBasedNNParams):
    """Config for transformer based log representation learning.

    :param nhead: number of attention heads.
    :param num_layers: number of hidden layers in the neural network.
    """

    nhead: int = 4
    num_layers: int = 1


class Transformer(ForecastBasedNN):
    """Transformer based model for learning log representation through a self-supervised forecasting task over
    log sequences.
    """

    def __init__(self, config: TransformerParams):
        super().__init__(config)
        self.config = config
        self.config.model_name = "transformer"
        num_labels = self.meta_data["num_labels"]
        # self.cls = torch.zeros(1, 1, self.config.embedding_dim).to(self.device)
        encoder_layer = nn.TransformerEncoderLayer(
            self.config.embedding_dim, self.config.nhead, self.config.hidden_size
        )
        self.transformer_encoder = nn.TransformerEncoder(
            encoder_layer, num_layers=self.config.num_layers
        )

        self.criterion = nn.CrossEntropyLoss()
        self.prediction_layer = nn.Linear(self.config.embedding_dim, num_labels)

    def forward(self, input_dict):
        """Forward method of transformer based model.

        :param input_dict: dict containing the session_idx, features, window_anomalies and window_labels as in ForecastNNVectorizedDataset object.
        :return: dict containing loss and prediction tensor.
        """
        if self.label_type == "anomaly":
            y = input_dict[ForecastNNVectorizedDataset.window_anomalies].long().view(-1)
        elif self.label_type == "next_log":
            y = input_dict[ForecastNNVectorizedDataset.window_labels].long().view(-1)
        self.batch_size = y.size()[0]
        x = input_dict[ForecastNNVectorizedDataset.features]
        x = self.embedder(x)

        x_t = x.transpose(1, 0)

        x_transformed = self.transformer_encoder(x_t.float())
        representation = x_transformed.transpose(1, 0).mean(dim=1)

        logits = self.prediction_layer(representation)
        y_pred = logits.softmax(dim=-1)
        loss = self.criterion(logits, y)
        return_dict = {"loss": loss, "y_pred": y_pred}
        return return_dict



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
import logging
import os
import pickle
from os.path import dirname, exists

import pandas as pd
import logai.algorithms.parsing_algo

from dataclasses import dataclass
from logai.config_interfaces import Config
from logai.utils import constants
from logai.algorithms.factory import factory


@dataclass
class LogParserConfig(Config):
    """
    Log Parser configuration.
    """

    parsing_algorithm: str = "drain"
    parsing_algo_params: object = None
    custom_config: object = None

    @classmethod
    def from_dict(cls, config_dict):
        config = super(LogParserConfig, cls).from_dict(config_dict)
        config.parsing_algo_params = factory.get_config(
            "parsing", config.parsing_algorithm.lower(), config.parsing_algo_params
        )
        return config


class LogParser:
    """
    Implementation of log parser for free-form text loglines.
    
    :param config: The log parser configuration.
    """

    def __init__(self, config: object):
        name = config.parsing_algorithm.lower()
        config_class = factory.get_config_class("parsing", name)
        algorithm_class = factory.get_algorithm_class("parsing", name)
        self.parser = algorithm_class(
            config.parsing_algo_params if config.parsing_algo_params else config_class()
        )

    def fit(self, loglines: pd.Series):
        """
        Trains log parser with training loglines.
        :param loglines: A pd.Series object containing the list of loglines for training.
        """
        self.parser.fit(loglines)

    def parse(self, loglines: pd.Series) -> pd.DataFrame:
        """
        Uses the trained log parser to parse loglines.
        :param loglines: A pd.Series object conatining the loglines for parsing.
        :return: A dataframe of parsed result ["loglines", "parsed_loglines", "parameter_list"].
        """
        if self.parser is None:
            raise RuntimeError("Parser is None.")
        parsed_loglines = self.parser.parse(loglines)
        if loglines.name is not constants.LOGLINE_NAME:
            loglines.name = constants.LOGLINE_NAME
        parsed_loglines.name = constants.PARSED_LOGLINE_NAME
        parsed_result = pd.concat([loglines, parsed_loglines], axis=1)

        parsed_result[constants.PARAMETER_LIST_NAME] = parsed_result.apply(
            self.get_parameter_list, axis=1
        )
        return parsed_result

    def fit_parse(self, loglines: pd.Series) -> pd.DataFrame:
        """
        Trains and parses the given loglines.
        :param loglines: A pd.Series object containing the list of loglines to train and parse.
        :return: A dataframe of parsed result ["loglines", "parsed_loglines", "parameter_list"].
        """
        try:
            self.fit(loglines)
        except RuntimeError as e:
            logging.ERROR("Cannot train parser")

        return self.parse(loglines)

    def save(self, out_path):
        """
        Saves the parser model.
        :param out_path: The directory to save parser models.
        """

        if not exists(dirname(out_path)):
            try:
                os.makedirs(dirname(out_path))
            except OSError as exc:
                if exc.errno != exc.errno.EEXIST:
                    raise RuntimeError(
                        "{} is not a valid output directory!".format(out_path)
                    )

        with open(out_path, "wb") as f:
            pickle.dump(self.parser, f)
            f.close()

    def load(self, model_path):
        """
        Loads existing parser models.
        :param model_path: The directory to load parser models.
        """

        with open(model_path, "rb") as f:
            self.parser = pickle.load(f)
            f.close()

    @staticmethod
    def get_parameter_list(row):
        """
        Returns parameter list of the loglines.

        :param row: The row in dataframe as function input containing ['logline', 'parsed_logline'].
        :return: The list of dynamic parameters.
        """
        parameter_list = []
        if not isinstance(row.logline, str) or not isinstance(row.parsed_logline, str):
            return parameter_list
        ll = row.logline.split()
        for t in ll:
            t = t.strip()
            if not t or t in row.parsed_logline:
                continue
            parameter_list.append(t)
        return parameter_list



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
from transformers import AutoTokenizer
import os


def get_tokenizer(tokenizer_dirpath):
    """Get huggingface tokenizer object from a given directory path.

    :param tokenizer_dirpath: (str): absolute path to directory containing pretrained tokenizer.
    :return:  AutoTokenizer: tokenizer object.
    """
    return AutoTokenizer.from_pretrained(tokenizer_dirpath, use_fast=True)


def get_special_tokens():
    """gets special tokens

    :return: list of special tokens
    """
    return [
        "[UNK]",
        "[PAD]",
        "[CLS]",
        "[SEP]",
        "[MASK]",
        ".",
        "*",
        ":",
        "$",
        "_",
        "-",
        "/",
    ]


def get_special_token_ids(tokenizer):
    """Get ids of special tokens, given a tokenizer object.

    :param tokenizer: (AutoTokenizer): tokenizer object.
    :return: list of token ids of special tokens.
    """
    return [tokenizer.convert_tokens_to_ids(x) for x in get_special_tokens()]


def get_tokenizer_vocab(tokenizer_dirpath):
    """Get vocabulary from a given tokenizer directory path.

    :param tokenizer_dirpath: (str): absolute path to directory containing pretrained tokenizer.
    :return: list of vocabulary words.
    """
    return open(os.path.join(tokenizer_dirpath, "vocab.txt")).readlines()


def get_mask_id(tokenizer):
    """Get id of mask token, given a tokenizer object.
    
    :param tokenizer: (AutoTokenizer): tokenizer object.
    :return: id of mask token.
    """
    return tokenizer.convert_tokens_to_ids("[MASK]")



# -*- coding: utf-8 -*-
#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#


# Learn more: https://github.com/kennethreitz/setup.py

from setuptools import setup, find_packages


with open("README.md") as f:
    readme = f.read()

with open("LICENSE.txt") as f:
    license = f.read()

extras_require = {
    "gui": [
        "dash-bootstrap-components>=1.2.1",
        "plotly>=5.9.0",
        "dash>=2.5.1",
    ],
    "deep-learning": [
        "tokenizers>=0.11.6",
        "datasets>=1.18.3",
        "torch>=1.10.1",
        "transformers>=4.17.0,<=4.23",
    ],
    "dev": [
        "Sphinx>=3.5.3",
        "docutils>=0.18.1",
        "wheel>=0.37.1",
        "toml>=0.10.2",
        "build>=0.7.0",
        "jupyter>=1.0.0",
        "ipykernel>=6.19.4",
        "pytest>=6.2.5",
    ]
}
extras_require["all"] = sum(extras_require.values(), [])

setup(
    name="logai",
    version="0.1.4",
    description="LogAI is unified framework for AI-based log analytics",
    long_description_content_type="text/markdown",
    long_description=readme,
    author="Qian Cheng, Amrita Saha, Wenzhuo Yang, Chenghao Liu, Gerald Woo, Doyen Sahoo, Steven Hoi",
    author_email="logai@salesforce.com",
    python_requires=">=3.8.0,<4",
    install_requires=[
        "scikit-learn>=1.2",
        "schema>=0.7.5",
        "salesforce-merlion>=1.0.0",
        "Cython>=0.29.30",
        "nltk>=3.6.5",
        "gensim>=4.1.2",
        "scikit-learn>=1.0.1",
        "pandas>=1.2.0",
        "numpy>=1.21.4",
        "spacy>=3.2.2",
        "attrs>=21.2.0",
        "dataclasses>=0.6",
        "PyYAML>=6.0",
        "tqdm>=4.62.3",
        "cachetools>=4.2.4",
        "matplotlib>=3.5.1",
        "seaborn>=0.11.2",
        "Jinja2>=3.0.0",
        "numba>=0.56.3",
    ],
    extras_require=extras_require,
    license=license,
    packages=find_packages(exclude=["tests", "tests.*", "docs", "gui", "gui.*"]),
    include_package_data=True,
)



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
import numpy as np
import pandas as pd
from attr import dataclass
from sklearn.cluster import Birch

from logai.algorithms.algo_interfaces import ClusteringAlgo
from logai.config_interfaces import Config
from logai.algorithms.factory import factory


@dataclass
class BirchParams(Config):
    """Parameters for Birch Clustering Algo. For more details on the parameters, see 
    https://scikit-learn.org/stable/modules/generated/sklearn.cluster.Birch.html.

    :param branching_factor: Maximum number of CF subclusters in each node.
    :param n_clusters: Number of clusters after the final clustering step, which treats the subclusters
        from the leaves as new samples.
    :param threshold: The radius of the subcluster obtained by merging a new sample and the closest
        subcluster should be lesser than the threshold.
    """
    branching_factor: int = 50
    n_clusters: int = None
    threshold: float = 1.5


@factory.register("clustering", "birch", BirchParams)
class BirchAlgo(ClusteringAlgo):
    """
    BIRCH algorithm for log clustering. This is a wrapper class for the Birch Clustering algorithm in scikit-learn 
    https://scikit-learn.org/stable/modules/generated/sklearn.cluster.Birch.html.
    """

    def __init__(self, params: BirchParams):
        self.model = Birch(
            branching_factor=params.branching_factor,
            n_clusters=params.n_clusters,
            threshold=params.threshold,
        )

    def fit(self, log_features: pd.DataFrame):
        """
        Trains a BIRCH model.

        :param log_features: The log features for training.
        """
        log_features = np.ascontiguousarray(log_features)
        self.model.partial_fit(log_features)

    def predict(self, log_features: pd.DataFrame) -> pd.Series:
        """
        Predicts using trained BIRCH model.

        :param log_features: The log features for inference.
        :return: A pandas series of log cluster labels.
        """
        log_features_carray = np.ascontiguousarray(log_features)

        res = self.model.predict(log_features_carray)
        return pd.Series(res, index=log_features.index)



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
from logai.preprocess.openset_preprocessor import OpenSetPreprocessor
from logai.preprocess.preprocessor import PreprocessorConfig
import pandas as pd
from logai.utils import constants
from logai.dataloader.data_model import LogRecordObject


class BGLPreprocessor(OpenSetPreprocessor):
    """
    Custom preprocessor for Open log dataset BGL.
    """

    def __init__(self, config: PreprocessorConfig):
        super().__init__(config)

    def _get_ids(self, logrecord: LogRecordObject) -> pd.Series:
        """Get ids of loglines.
        
        :param logrecord:  logrecord object containing the BGL data.
        :return: pd.Series object containing the ids of the loglines.
        """
        time_unit_in_secs = 60  # 21600.0 # 6 hours
        ids = logrecord.span_id[constants.SPAN_ID].astype(int)
        start_time = ids[0]
        session_ids = ids.apply(lambda x: int((x - start_time) / time_unit_in_secs))
        return session_ids

    def _get_labels(self, logrecord: LogRecordObject):
        """Get anomaly detection labels of loglines.
        
        :param logrecord: logrecord object containing the BGL data.
        :return:pd.Series object containing the labels of the loglines.
        """
        return logrecord.labels[constants.LABELS].apply(lambda x: int(x != "-"))



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
import abc

import pandas as pd
from logai.dataloader.data_model import LogRecordObject


class ParsingAlgo(abc.ABC):
    """
    Interface for parsing algorithms.

    """

    @abc.abstractmethod
    def fit(self, loglines: pd.Series):
        """
        fit parsing algorithm with input

        :param loglines: pd.Series of loglines as input
        :return: pd.Dataframe

        """
        pass

    @abc.abstractmethod
    def parse(self, loglines: pd.Series) -> pd.DataFrame:
        """
        parse loglines
        :param loglines: pd.Series of loglines to parse
        :return: pd.Dataframe of parsed results ["loglines", "parsed_loglines", "parameter_list"].
        """
        pass


class VectorizationAlgo(abc.ABC):
    """
    Interface for logline vectorization algorithms
    """

    @abc.abstractmethod
    def fit(self, loglines: pd.Series):
        """
        fit vectorizor with input.
        :param loglines:
        :return:
        """
        pass

    @abc.abstractmethod
    def transform(self, loglines: pd.Series) -> pd.DataFrame:
        """
        transform given loglines into vectors.
        :param loglines:
        :return:
        """
        pass


class FeatureExtractionAlgo(abc.ABC):
    """
    Interface for feature extraction algorithms
    """

    @abc.abstractmethod
    def extract(self):
        pass


class ClusteringAlgo(abc.ABC):
    """
    Interface for clustering algorithms
    """

    @abc.abstractmethod
    def fit(self, log_features: pd.DataFrame):
        pass

    @abc.abstractmethod
    def predict(self, log_features: pd.DataFrame):
        pass


class AnomalyDetectionAlgo(abc.ABC):
    """
    Interface for clustering algorithms
    """

    @abc.abstractmethod
    def fit(self, log_features: pd.DataFrame):
        pass

    @abc.abstractmethod
    def predict(self, log_features: pd.DataFrame) -> pd.Series:
        pass


class NNAnomalyDetectionAlgo(abc.ABC):
    @abc.abstractmethod
    def fit(self, train_data, dev_data: LogRecordObject):
        pass

    @abc.abstractmethod
    def predict(self, test_data: LogRecordObject):
        pass


class CategoricalEncodingAlgo(abc.ABC):
    """
    Interface for categorical encoders
    """

    @abc.abstractmethod
    def fit_transform(self, log_attributes: pd.DataFrame) -> pd.DataFrame:
        pass



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
import pandas as pd
import logai.algorithms.anomaly_detection_algo

from attr import dataclass
from logai.config_interfaces import Config
from logai.algorithms.factory import factory


@dataclass
class AnomalyDetectionConfig(Config):
    """Config class for AnomalyDetector.

    :param algo_name: The algorithm name.
    :param algo_params: The algorithm parameters.
    :param custom_params: Additional customized parameters.
    """
    algo_name: str = "one_class_svm"
    algo_params: object = None
    custom_params: object = None

    @classmethod
    def from_dict(cls, config_dict):
        config = super(AnomalyDetectionConfig, cls).from_dict(config_dict)
        config.algo_params = factory.get_config(
            "detection", config.algo_name.lower(), config.algo_params
        )
        return config


class AnomalyDetector:
    def __init__(self, config: AnomalyDetectionConfig):
        """
        Initializes the anomaly detector with proper configuration. If no config provided, use default.

        :param config: A config object for anomaly detection.
        """
        self.anomaly_detector = factory.get_algorithm(
            "detection", config.algo_name.lower(), config
        )

    def fit(self, log_features: pd.DataFrame):
        """
        Trains an anomaly detection given the training dataset.

        :param log_features: The training dataset.
        """
        return self.anomaly_detector.fit(log_features)

    def predict(self, log_features: pd.DataFrame) -> pd.DataFrame:
        """
        Predicts anomalies given the test dataset.

        :param log_features: The test dataset.
        :return: A pandas dataframe containing the prediction results.
        """
        return self.anomaly_detector.predict(log_features)



#
# Copyright (c) 2023 Salesforce.com, inc.
# All rights reserved.
# SPDX-License-Identifier: BSD-3-Clause
# For full license text, see the LICENSE file in the repo root or https://opensource.org/licenses/BSD-3-Clause
#
#
import numpy as np
import pandas as pd
from dataclasses import dataclass, fields, _MISSING_TYPE

from logai.config_interfaces import Config
from logai.utils import constants
from logai.utils.functions import pad


@dataclass
class FeatureExtractorConfig(Config):
    """Config class for Feature Extractor.
    
    :param group_by_category: Which fields of the dataframe object to group by.
    :param group_by_time: Grouping log lines by the time frequency, using the notations in
        https://pandas.pydata.org/pandas-docs/stable/user_guide/timeseries.html#offset-aliases.
    :param sliding_window: The length of the sliding window .
    :param steps: The step-size of sliding window.
    :param max_feature_len: The pad the log vector to this size.
    """
    group_by_category: list = None
    group_by_time: str = None
    sliding_window: int = 0
    steps: int = 1
    max_feature_len: int = 100

    def __post_init__(self):
        # Loop through the fields
        for field in fields(self):
            # If there is a default and the value of the field is none we can assign a value
            if (
                not isinstance(field.default, _MISSING_TYPE)
                and getattr(self, field.name) is None
            ):
                setattr(self, field.name, field.default)


def _get_group_counter(attributes: pd.DataFrame, group_by_category: list) -> pd.Series:
    """
    Merge counting with other feature extraction functions.
    
    :param attributes: pd.Dataframe or pd.Series for counting.
    :param group_by_category: selected attributes for grouping and counting.
    """
    attributes["group_index"] = attributes.index
    group_counter_list = (
        attributes.groupby(by=group_by_category)
        .size()
        .rename(constants.LOG_COUNTS)
        .reset_index()
    )
    return group_counter_list


class FeatureExtractor:
    """
    Feature Extractor combines the structured log attributes and log vectors. Feature extractor can group
    log records to log events based on user defined strategies. Such as group by a categorical column, or group by
    timestamps.
    Generate feature sets:
    1. log features: generate feature set from log vectors.
    2. log event sequence: concatenating all loglines belongs to the same log event.
    3. log event counter vector: for each log event.
    4. log vector
    Note:
        1. counter vector
        2. sematic vector
        3. id sequence vector.

        partitioning:
        1. group by attributes
        2. group by length of sequence, either sliding windows or fix window.
        3. timestamp interval based.
    """

    def __init__(self, config: FeatureExtractorConfig):
        self.config = config

    def convert_to_counter_vector(
        self,
        log_pattern: pd.Series = None,
        attributes: pd.DataFrame = None,
        timestamps: pd.Series = None,
    ) -> pd.DataFrame:
        """
        Converts logs to log counter vector, after grouping log data based on the FeatureExtractor config.
        
        :param log_pattern: The unstructured part of the log data.
        :param attributes: The log attributes.
        :param timestamps: The timestamps.

        :return: The dataframe object containing the counts of the log-events after grouping.
        """
        # TODO: Implement sliding window for counter vectors
        input_df = self._get_input_df(log_pattern, attributes, timestamps)

        gb = self._get_group(input_df)

        event_index_list = self._get_group_index(gb)
        event_index_list[constants.LOG_COUNTS] = event_index_list["event_index"].apply(
            lambda x: len(x)
        )
        return event_index_list

    def convert_to_feature_vector(
        self,
        log_vectors: pd.Series,
        attributes: pd.DataFrame,
        timestamps: pd.Series,
    ) -> pd.DataFrame:
        """
        Converts log data into feature vector, by combining the log vectors (can be output
        of LogVectorizer) with other numerical or categorical attributes of the logs,
        after grouping based on the FeatureExtractorConfig.
        
        :param log_vectors: Numeric features of the logs (for e.g. the vectorized form of the log data obtained as output of LogVectorizer).
        :param attributes: Categorical or numerical attributes for grouping, or numerical attributes serve as additional features.
        :param timestamps: pd.Series object containing the timestamp data of the loglines.

        :return: ``event_index_list``: modified log data (pd.DataFrame) consisting of the converted feature vector form of the input log data
            after applying the log grouping. It contains an "event_index" field which maintains the sequence of
            log event ids where these ids correspond to the  original input dataframe's indices.
            ``block_list``: pd.DataFrame object.
        """
        if log_vectors is None or log_vectors.empty:
            feature_df = None
        else:
            feature_df = self._convert_to_feature_df(log_vectors)
        input_df = self._get_input_df(feature_df, attributes, timestamps)

        gb = self._get_group(input_df)

        # TODO: currently using mean() to aggregate log features. This may be expanded to take any stats.
        # TODO: implement sliding window for feature vectors.
        block_list = gb.mean().reset_index()

        event_index_list = gb.agg(list).reset_index()

        return event_index_list, block_list.loc[:, block_list.columns != "event_index"]

    def convert_to_sequence(
        self,
        log_pattern: pd.Series = None,
        attributes: pd.DataFrame = None,
        timestamps: pd.Series = None,
    ):
        """Converts log data into sequence using sliding window technique, as defined in FeatureExtractorConfig.
        
        :param log_pattern: A pd.Series object that encapsulates the entire arbitrary unstructured part of the log data
            (for example, can be the unstructured part of the raw log data or the output of the output of the log parser).
        :param attributes: The structured part (attributes) of the raw log data.
        :param timestamps: The timestamps data corresponding to the log lines.
        :return: ``event_index_list``: pd.DataFrame object of modified log data consisting of the sequence form of the
            structured and unstructured input data (i.e. log_pattern and attributes arguments) after running sliding
            window. For the unstructured part, the returned DataFrame contains an "event_index" field which maintains
            the sequence of log event ids where these ids correspond to the original input dataframe's indices.
            ``event_sequence``: pd.Series object containing the concatenating form of the unstructured input data
            (i.e. log_pattern argument), after concatenating the unstructured data for each sliding window.
        """
        # TODO: Converting sequence by sliding windows.
        # Partioning: length of sequence, step
        # Attributes: session id.
        input_df = self._get_input_df(log_pattern, attributes, timestamps)

        gb = self._get_group(input_df)

        if self.config.sliding_window > 0:
            if self.config.steps <= 0:
                raise RuntimeError(
                    "Step should be greater than zero. Step: {}".format(
                        self.config.steps
                    )
                )
            window = self.config.sliding_window
            step = self.config.steps
            raw_index_list = gb.agg(list).reset_index()
            colnames = raw_index_list.columns

            event_index_list = pd.DataFrame(columns=colnames)
            for ind, row in raw_index_list.iterrows():
                indices = row["event_index"]
                loglines = row[log_pattern.name]
                if len(indices) < self.config.sliding_window:
                    event_index_list = event_index_list.append(row)
                else:
                    event_indices = np.lib.stride_tricks.sliding_window_view(
                        indices, window
                    )[::step, :].tolist()
                    log_seq = np.lib.stride_tricks.sliding_window_view(
                        loglines, window
                    )[::step, :].tolist()
                    # when timestamps not a groupby factor, we need to partition timestamps.
                    if (not timestamps.empty) and (not self.config.group_by_time):
                        ts = row[constants.LOG_TIMESTAMPS]
                        event_ts = np.lib.stride_tricks.sliding_window_view(ts, window)[
                            ::step, :
                        ].tolist()
                    for i in range(len(event_indices)):
                        row["event_index"] = event_indices[i]
                        row[log_pattern.name] = log_seq[i]
                        if (not timestamps.empty) and (not self.config.group_by_time):
                            row[constants.LOG_TIMESTAMPS] = event_ts[i]
                        event_index_list = event_index_list.append(row)
            event_sequence = event_index_list[log_pattern.name].apply(
                lambda x: " ".join(x)
            )

        else:
            event_index_list = gb.agg(list).reset_index()
            event_sequence = event_index_list["event_index"].apply(
                lambda x: " ".join([str(log_pattern[c]) for c in x])
            )

        return event_index_list, event_sequence

    def _get_input_df(
        self,
        logline: pd.Series or pd.DataFrame,
        attributes: pd.DataFrame,
        timestamps: pd.Series,
    ) -> pd.DataFrame:
        if timestamps is not None:
            timestamps = timestamps.rename(constants.LOG_TIMESTAMPS)

        if logline is None:
            input_df = pd.concat((attributes, timestamps), axis=1)
        else:
            input_df = pd.concat((logline, attributes, timestamps), axis=1)
        input_df["event_index"] = input_df.index.values

        return input_df

    def _get_group_index(self, gb) -> pd.Series:
        event_index_list = gb.agg(list).reset_index()
        # print(event_index_list.columns)

        return event_index_list

    def _get_group_size(self, event_index_list) -> pd.Series:

        size_list = event_index_list["event_index_list"].apply(lambda x: len(x))
        return size_list

    def _get_group(self, input_df):
        group_by = []
        if self.config.group_by_category:
            group_by += self.config.group_by_category

        if self.config.group_by_time:
            input_df[constants.LOG_TIMESTAMPS] = input_df[
                constants.LOG_TIMESTAMPS
            ].dt.floor(freq=self.config.group_by_time)
            group_by += [constants.LOG_TIMESTAMPS]
        if group_by:
            return input_df.groupby(by=group_by)
        else:
            return input_df.groupby(by=["event_index"])

    def _convert_to_feature_df(self, log_vectors: pd.Series) -> pd.DataFrame:
        padded_log_vectors = log_vectors.apply(pad, max_len=self.config.max_feature_len)
        log_features = pd.DataFrame(
            padded_log_vectors.tolist(), index=log_vectors.index
        )
        log_features.columns = [
            "feature_{}".format(i) for i in range(len(log_features.columns))
        ]
        return log_features
